import { isDev } from '../../core/env';
import ContentSecurityPolicy = Session.ContentSecurityPolicy;

const defaultCsp: Session.ContentSecurityPolicy = {
  'default-src': ["'self'"],
  'style-src-attr': ["'unsafe-inline'"],
};

function stringifyCspHeader(csp: ContentSecurityPolicy): string {
  return Object.entries(csp)
    .map(
      (directive: [string, Session.CspValue[] | undefined]) =>
        `${directive[0]} ${directive[1] ? directive[1]?.join(' ') : ''}`,
    )
    .join('; ');
}

export function mergeContentSecurityPolicy(...contentSecurityPolicies: ContentSecurityPolicy[]): ContentSecurityPolicy {
  return contentSecurityPolicies.reduce((mergedCsp, contentSecurityPolicy) => {
    Object.entries(contentSecurityPolicy).forEach(([policyName, policy]) => {
      const mergedPolicy = mergedCsp[policyName as keyof ContentSecurityPolicy];
      if (mergedPolicy && policy) {
        mergedPolicy.push(...policy);
      } else if (policy) {
        mergedCsp[policyName as keyof ContentSecurityPolicy] = policy;
      }
    });
    return mergedCsp;
  }, {});
}

export function setWindowCsp(window: Electron.BrowserWindow, csp: ContentSecurityPolicy): void {
  const mergedCsp: ContentSecurityPolicy = { ...defaultCsp, ...csp };

  if (isDev()) {
    mergedCsp['default-src'] = ['devtools:'].concat(mergedCsp['default-src'] ?? []);
    mergedCsp['script-src'] = ["'unsafe-eval'"].concat(mergedCsp['script-src'] ?? []);
    mergedCsp['script-src-elem'] = ['file:', 'devtools:', "'unsafe-inline'"].concat(mergedCsp['script-src-elem'] ?? []);
    mergedCsp['style-src'] = ['devtools:', "'unsafe-inline'"].concat(mergedCsp['style-src'] ?? []);
    mergedCsp['img-src'] = ['devtools:'].concat(mergedCsp['img-src'] ?? []);
    mergedCsp['connect-src'] = ['devtools:', 'data:'].concat(mergedCsp['connect-src'] ?? []);
    mergedCsp['worker-src'] = ['devtools:'].concat(mergedCsp['worker-src'] ?? []);
  }

  window.webContents.session.webRequest.onHeadersReceived((details, callback) => {
    callback({
      responseHeaders: {
        ...details.responseHeaders,
        'Content-Security-Policy': stringifyCspHeader(mergedCsp),
      },
    });
  });
}
